from enum import Enum


class Hook(Enum):
    JSON_FINALIZER = "__finalize_json__"
    ARGS_PREPARER = "__prepare_args__"

    QUERY_CASE = "__query_case__"
    BODY_CASE = "__body_case__"
    COOKIE_CASE = "__cookie_case__"
    HEADER_CASE = "__header_case__"
    RESPONSE_CASE = "__response_case__"

    @classmethod
    def values(cls) -> list[str]:
        return [member.value for member in cls]



import re

__all__ = [
    'snake_case',
    'camel_case',
    'pascal_case',
    'constant_case',
    'kebab_case',
    'header_case'
]


def snake_case(s: str) -> str:
    """Convert a string to snake_case. """
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)
    s = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s)
    s = re.sub(r'\W+', '_', s).lower()
    s = re.sub(r'_+', '_', s)
    return s


def camel_case(s: str) -> str:
    """Convert a string to camelCase."""
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)
    s = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s)
    s = re.sub(r'\W+', '_', s)
    words = s.split('_')
    capitalized_words = [word.capitalize() for word in words]
    return capitalized_words[0].lower() + ''.join(capitalized_words[1:])


def pascal_case(s: str) -> str:
    """Convert a string to PascalCase."""
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)
    s = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s)
    s = re.sub(r'\W+', '_', s)
    words = s.split('_')
    capitalized_words = [word.capitalize() for word in words]
    return ''.join(capitalized_words)


def constant_case(s: str) -> str:
    """Convert a string to CONSTANT_CASE."""
    s = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', s)
    s = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', s)
    s = re.sub(r'[\W_]+', '_', s)
    return s.upper()


def kebab_case(s: str) -> str:
    """Convert a string to kebab-case."""
    s = re.sub(r"(\s|_|-)+", " ", s)
    s = re.sub(r"[A-Z]{2,}(?=[A-Z][a-z]+[0-9]*|\b)|[A-Z]?[a-z]+[0-9]*|[A-Z]|[0-9]+",
               lambda mo: ' ' + mo.group(0).lower(), s)
    s = '-'.join(s.split())
    return s


def header_case(s: str) -> str:
    """Convert a string to Header-Case."""
    s = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', s)
    s = re.sub('([a-z0-9])([A-Z])', r'\1 \2', s)
    s = re.sub(r'[_\W]+', ' ', s)
    words = s.split()
    capitalized_words = [word.capitalize() for word in words]
    return '-'.join(capitalized_words)



from __future__ import annotations

from functools import partial
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo
from sensei.types import IResponse, Json
from sensei._utils import format_str
from sensei.params import Body, Query, Header, Cookie
from sensei.cases import header_case as to_header_case
from typing import Any, get_args, Callable, Annotated, TypeVar, Generic, get_origin
from ..tools import ChainedMap, split_params, make_model, is_safe_method, HTTPMethod, validate_method, identical

CaseConverter = Callable[[str], str]

ResponseModel = TypeVar(
    'ResponseModel',
    type[BaseModel],
    str,
    dict,
    bytes,
    list[dict],
    BaseModel,
    list[BaseModel],
)

RESPONSE_TYPES = ResponseModel.__constraints__
_ConditionChecker = Callable[[type[ResponseModel]], bool]
_ResponseHandler = Callable[[type[ResponseModel], IResponse], ResponseModel]
_PartialHandler = Callable[[IResponse], ResponseModel]


class Args(BaseModel):
    """
    Model used in preparers as input and output argument. Stores request arguments

    Attributes:
        url (str): URL to which the request will be made.
        params (dict[str, Any]): Dictionary of query parameters to be included in the URL.
                                 Default is an empty dictionary.
        json_ (Json): JSON payload for the request body.
                                The field is aliased as 'json' and defaults to an empty dictionary.
        headers (dict[str, Any]): Dictionary of HTTP headers to be sent with the request.
                                  Default is an empty dictionary.
        cookies (dict[str, Any]): Dictionary of cookies to be included in the request.
                                  Default is an empty dictionary.
    """
    url: str
    params: dict[str, Any] = {}
    json_: Json = Field({}, alias="json")
    headers: dict[str, Any] = {}
    cookies: dict[str, Any] = {}


class Endpoint(Generic[ResponseModel]):
    __slots__ = (
        "_path",
        "_method",
        "_error_msg",
        "_case_converters",
        "_params_model",
        "_response_model",
    )

    _response_handle_map: dict[_ConditionChecker, _ResponseHandler] = {
        lambda model: isinstance(model, type(BaseModel)): lambda model, response: model(**response.json()),
        lambda model: model is str: lambda model, response: response.text,
        lambda model: dict in (model, get_origin(model)): lambda model, response: response.json(),
        lambda model: model is bytes: lambda model, response: response.content,
        lambda model: isinstance(model, BaseModel): lambda model, response: model,
        lambda model: model is None: lambda model, response: None,
        lambda model: get_origin(model) is list and isinstance(get_args(model)[0], type(BaseModel)):
            lambda model, response: [get_args(model)[0](**value) for value in response.json()],
        lambda model: (get_origin(model) is list and ((arg := get_args(model)[0]) is dict or get_origin(arg) is dict)):
            lambda model, response: response.json(),
    }

    def __init__(
            self,
            path: str,
            method: HTTPMethod,
            /, *,
            params: dict[str, Any] | None = None,
            response: type[ResponseModel] | None = None,
            error_msg: str | None = None,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = to_header_case
    ):
        validate_method(method)

        self._path = path
        self._method = method
        self._error_msg = error_msg

        params_model = self._make_model('Params', params)

        converters = {
            'params': query_case,
            'json': body_case,
            'cookies': cookie_case,
            'headers': header_case
        }

        for k in converters.keys():
            if converters[k] is None:
                converters[k] = identical

        self._case_converters = converters

        self._params_model = params_model
        self._response_model = response

    @property
    def path(self) -> str:
        return self._path

    @property
    def method(self) -> HTTPMethod:
        return self._method

    @property
    def params_model(self) -> type[BaseModel] | None:
        return self._params_model

    @property
    def response_model(self) -> type[ResponseModel] | None:
        return self._response_model

    @staticmethod
    def _make_model(model_name: str, model_args: dict[str, Any] | None) -> type[BaseModel] | None:
        if model_args:
            return make_model(model_name, model_args)
        else:
            return None

    @classmethod
    def _handle_if_condition(cls, model: type[ResponseModel]) -> _PartialHandler:
        for checker, handler in cls._response_handle_map.items():
            if checker(model):
                result = partial(handler, model)
                break
        else:
            raise ValueError('Unsupported response type')

        return result

    @classmethod
    def is_response_type(cls, value: Any) -> bool:
        try:
            cls._handle_if_condition(value)
            return True
        except ValueError:
            return False

    def get_args(self, **kwargs) -> Args:
        params_model = self.params_model
        path = self.path
        if params_model:
            url, request_params = self._get_init_args(params_model, **kwargs)
        else:
            url = path
            request_params = {}

        return Args(
            url=url,
            **request_params
        )

    def get_response(self, *, response_obj: IResponse) -> ResponseModel | None:
        response_model = self.response_model

        result = self._handle_if_condition(response_model)(response_obj)

        to_validate = result

        if isinstance(response_model, BaseModel):
            response_model = type(response_model)
            to_validate = response_model(**result.model_dump(mode='json', by_alias=True))

        class ValidationModel(BaseModel):
            result: response_model

        ValidationModel(result=to_validate)

        return result

    def _get_init_args(
            self,
            params_model: type[BaseModel],
            **kwargs
    ) -> tuple[str, dict[str, Any]]:
        path = self.path
        params_model_instance = params_model(**kwargs)
        params_all = params_model_instance.model_dump(mode='json', by_alias=True)
        params, path_params = split_params(path, params_all)

        annotations_all = params_model.__annotations__.copy()
        annotations, _ = split_params(path, annotations_all)

        request_params = self._map_params(annotations, params)

        url = format_str(path, path_params, True)
        return url, request_params

    def _map_params(
            self,
            annotations: dict[str, Any],
            params: dict[str, Any]
    ) -> dict[str, Any]:
        new_params = {
            'params': {},
            'json': {},
            'headers': {},
            'cookies': {}
        }

        annotation_map = {
            Query: 'params',
            Body: 'json',
            Cookie: 'cookies',
            Header: 'headers',
        }

        type_to_converter = ChainedMap[type[FieldInfo], CaseConverter](annotation_map, self._case_converters)
        type_to_params = ChainedMap[type[FieldInfo], dict[str, Any]](annotation_map, new_params)
        temp_type = type(Annotated[int, int])

        for key, value in annotations.items():
            if isinstance(value, temp_type):
                param_annotation = get_args(value)[1]

                param_type = type(param_annotation)

                converted = type_to_converter[param_type](key)

                result_key = param_annotation.alias if param_annotation.alias else converted
                type_to_params[param_type][result_key] = params[key]
            else:
                param_type = Query if is_safe_method(self.method) else Body

                converted = type_to_converter[param_type](key)
                type_to_params[param_type][converted] = params[key]

        new_params = {k: v for k, v in new_params.items() if v}
        return new_params



from __future__ import annotations
from functools import wraps
from httpx import URL
from typing import Callable, Optional, Any
from sensei.client import Manager
from ._endpoint import CaseConverter
from ._requester import JsonFinalizer, Preparer
from ._route import Route
from ..tools import HTTPMethod, set_method_type, identical, MethodType
from sensei._utils import bind_attributes, get_base_url
from ._types import RoutedFunction, IRouter, RoutedModel
from ._hook import Hook
from sensei.types import IRateLimit
from sensei._descriptors import RateLimitAttr, PortAttr
from sensei.cases import header_case as to_header_case

_RouteDecorator = Callable[[Callable], RoutedFunction]


class Router(IRouter):
    """
    Router for managing API routes and handling HTTP requests.

    The `Router` class provides decorators for defining API endpoints with various HTTP methods
    (GET, POST, PATCH, PUT, DELETE). It handles URL construction, request preparation, JSON finalization,
    and rate limiting.

    Args:
        host (str):
            The root URL of the associated API. It may contain a colon and a placeholder for the port, e.g., `:{port}`.
            If a port is provided, it will replace the placeholder.
        port (Optional[int], optional):
            The port number of the associated API. If `None`, the port placeholder in `host` will not be replaced.
            Defaults to `None`.
        rate_limit (Optional[IRateLimit], optional):
            An object implementing the `IRateLimit` interface to handle API rate limiting.
            Defaults to `None`.
        manager (Optional[Manager], optional):
            A `Manager` instance used to provide an HTTP client to the router.
            Defaults to `None`.
        query_case (CaseConverter, optional):
            A converter function for the case of query parameters.
            Defaults to `identical`.
        body_case (CaseConverter, optional):
            A converter function for the case of body parameters.
            Defaults to `identical`.
        cookie_case (CaseConverter, optional):
            A converter function for the case of cookies.
            Defaults to `identical`.
        header_case (CaseConverter, optional):
            A converter function for the case of headers.
            Defaults to `identical`.
        __finalize_json__ (JsonFinalizer, optional):
            A function to finalize the JSON response. The final value must be JSON serializable.
            Defaults to `identical`.
        __prepare_args__ (Preparer, optional):
            A preparer function used to prepare the arguments for the request before it is sent.
            The final value must be an instance of `Args`. Defaults to `identical`.
    """

    rate_limit = RateLimitAttr()
    port = PortAttr()

    def __init__(
            self,
            host: str,
            *,
            port: Optional[int] = None,
            rate_limit: Optional[IRateLimit] = None,
            manager: Optional[Manager] = None,
            query_case: CaseConverter = identical,
            body_case: CaseConverter = identical,
            cookie_case: CaseConverter = identical,
            header_case: CaseConverter = to_header_case,
            response_case: CaseConverter = identical,
            __finalize_json__: JsonFinalizer = identical,
            __prepare_args__: Preparer = identical
    ):
        self._manager = manager
        self._host = host

        self.port = port
        self.rate_limit = rate_limit

        self._query_case = query_case
        self._body_case = body_case
        self._cookie_case = cookie_case
        self._header_case = header_case
        self._response_case = response_case

        self._finalize_json = __finalize_json__
        self._prepare_args = __prepare_args__
        self._linked_to_model: bool = False

    @property
    def base_url(self) -> URL:
        """
        Get the base URL constructed from the host and port.

        Returns:
            str: The base URL.
        """
        host = self._host
        port = self.port
        base = get_base_url(host, port)
        return URL(base)

    @property
    def manager(self) -> Optional[Manager]:
        """
        Get the manager used to provide an HTTP client to the router.

        Returns:
            Optional[Manager]: The `Manager` instance if set; otherwise, `None`.
        """
        return self._manager

    def _replace_default_converters(
            self,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            response_case: CaseConverter | None = None,
    ) -> dict[str, CaseConverter]:
        """
        Replace default case converters with provided ones if any.

        Args:
            query_case (CaseConverter | None, optional):
                Case converter for query parameters. Defaults to `None`.
            body_case (CaseConverter | None, optional):
                Case converter for body parameters. Defaults to `None`.
            cookie_case (CaseConverter | None, optional):
                Case converter for cookies. Defaults to `None`.
            header_case (CaseConverter | None, optional):
                Case converter for headers. Defaults to `None`.

        Returns:
            dict[str, CaseConverter]: A dictionary mapping parameter types to their converters.
        """
        converters = {
            'query_case': query_case,
            'body_case': body_case,
            'cookie_case': cookie_case,
            'header_case': header_case,
            'response_case': response_case,
        }

        for key, converter in converters.items():
            if converter is None:
                converters[key] = getattr(self, f'_{key}')

        return converters

    def _get_decorator(
            self,
            path: str,
            method: HTTPMethod,
            *,
            case_converters: dict[str, CaseConverter],
            skip_finalizer: bool = False,
    ) -> Callable:
        """
        Create a decorator for a specific HTTP method and path.

        Args:
            path (str):
                The relative path of the route.
            method (HTTPMethod):
                The HTTP method for the route (e.g., GET, POST).
            case_converters (dict[str, CaseConverter]):
                A dictionary of case converters for query, body, cookie, and header parameters.
            skip_finalizer:
                Skip JSON finalizer, when finalizing response. Defaults to `False`.

        Returns:
            Callable: A decorator that transforms a function into a routed function.
        """

        def decorator(func: Callable) -> Callable:
            json_finalizer = identical if skip_finalizer else self._finalize_json
            route = Route(
                path=path,
                method=method,
                func=func,
                manager=self._manager,
                host=self._host,
                port=self.port,
                rate_limit=self.rate_limit,
                case_converters=case_converters,
                json_finalizer=json_finalizer,
                pre_preparer=self._prepare_args
            )

            def _setattrs(
                    instance: Any,
                    func: Callable,
                    wrapper: Callable,
                    route: Route
            ) -> None:
                method_type = route.method_type = wrapper.__method_type__  # type: ignore
                if MethodType.self_method(method_type):
                    route.__self__ = instance
                    func.__self__ = instance

            if not route.is_async:
                @set_method_type
                @wraps(func)
                def wrapper(*args, **kwargs):
                    instance = args[0] if len(args) else None
                    _setattrs(instance, func, wrapper, route)
                    res = route(*args, **kwargs)
                    _setattrs(None, func, wrapper, route)
                    return res
            else:
                @set_method_type
                @wraps(func)
                async def wrapper(*args, **kwargs):
                    instance = args[0] if len(args) else None
                    _setattrs(instance, func, wrapper, route)
                    res = await route(*args, **kwargs)
                    _setattrs(None, func, wrapper, route)
                    return res

            bind_attributes(wrapper, route.finalize, route.prepare)  # type: ignore
            return wrapper

        return decorator

    def model(self, model_cls: Optional[RoutedModel] = None) -> RoutedModel:
        """
        Associate a model class with the router to enable "routed" methods and `dunder` hooks.

        This method modifies a class derived from `APIModel` to work seamlessly with the router,
        allowing the use of routed methods and special hooks.

        Args:
            model_cls (Optional[RoutedModel], optional):
                The class to be modified. If `None`, returns a decorator for later use.
                Defaults to `None`.

        Raises:
            ValueError: If a model is already linked to the router.
        """

        def decorator(model_cls: RoutedModel) -> RoutedModel:
            if self._linked_to_model:
                raise ValueError('Only one model can be associated with a router')

            model_cls.__router__ = self
            self._linked_to_model = True

            hooks = Hook.values()

            for hook in hooks:
                hook_fun = getattr(model_cls, hook, None)
                if hook_fun:
                    setattr(self, hook[1:-2], hook_fun)

            return model_cls

        if model_cls is None:
            return decorator  # type: ignore
        else:
            return decorator(model_cls)

    def get(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            response_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> _RouteDecorator:
        """
        Create a route using the HTTP GET method.

        This decorator transforms a function into a routed function that sends a GET request
        to the specified path, handling parameter case conversion and argument validation.

        Args:
            path (str):
                The relative path of the route.
            query_case (CaseConverter | None, optional):
                Case converter for query parameters. Defaults to using the router's default.
            body_case (CaseConverter | None, optional):
                Case converter for body parameters. Defaults to using the router's default.
            cookie_case (CaseConverter | None, optional):
                Case converter for cookie parameters. Defaults to using the router's default.
            header_case (CaseConverter | None, optional):
                Case converter for header parameters. Defaults to using the router's default.
            response_case (CaseConverter | None, optional):
                Case converter for JSON response. Defaults to using the router's default.
            skip_finalizer:
                Skip JSON finalizer, when finalizing response. Defaults to `False`.

        Returns:
            RoutedFunction: A routed function that sends a GET request according to the specified path and validates
            its arguments based on type annotations.

        Raises:
            pydantic_core.ValidationError: If type validation of arguments fails.
        """
        converters = self._replace_default_converters(query_case, body_case, cookie_case, header_case, response_case)

        decorator = self._get_decorator(
            path=path,
            method="GET",
            case_converters=converters,
            skip_finalizer=skip_finalizer
        )
        return decorator

    def post(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            response_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> _RouteDecorator:
        """
        Create a route using the HTTP POST method.

        This decorator transforms a function into a routed function that sends a POST request
        to the specified path, handling parameter case conversion and argument validation.

        Args:
            path (str):
                The relative path of the route.
            query_case (CaseConverter | None, optional):
                Case converter for query parameters. Defaults to using the router's default.
            body_case (CaseConverter | None, optional):
                Case converter for body parameters. Defaults to using the router's default.
            cookie_case (CaseConverter | None, optional):
                Case converter for cookie parameters. Defaults to using the router's default.
            header_case (CaseConverter | None, optional):
                Case converter for header parameters. Defaults to using the router's default.
            response_case:
                Case converter for JSON response. Defaults to using the router's default.
            skip_finalizer:
                Skip JSON finalizer, when finalizing response. Defaults to `False`.

        Returns:
            RoutedFunction: A routed function that sends a POST request according to the specified path and validates
            its arguments based on type annotations.

        Raises:
            pydantic_core.ValidationError: If type validation of arguments fails.
        """
        converters = self._replace_default_converters(query_case, body_case, cookie_case, header_case, response_case)

        decorator = self._get_decorator(
            path=path,
            method="POST",
            case_converters=converters,
            skip_finalizer=skip_finalizer
        )
        return decorator

    def patch(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            response_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> _RouteDecorator:
        """
        Create a route using the HTTP PATCH method.

        This decorator transforms a function into a routed function that sends a PATCH request
        to the specified path, handling parameter case conversion and argument validation.

        Args:
            path (str):
                The relative path of the route.
            query_case (CaseConverter | None, optional):
                Case converter for query parameters. Defaults to using the router's default.
            body_case (CaseConverter | None, optional):
                Case converter for body parameters. Defaults to using the router's default.
            cookie_case (CaseConverter | None, optional):
                Case converter for cookie parameters. Defaults to using the router's default.
            header_case (CaseConverter | None, optional):
                Case converter for header parameters. Defaults to using the router's default.
            response_case:
                Case converter for JSON response. Defaults to using the router's default.
            skip_finalizer:
                Skip JSON finalizer, when finalizing response. Defaults to `False`.

        Returns:
            RoutedFunction: A routed function that sends a PATCH request according to the specified path and validates
            its arguments based on type annotations.

        Raises:
            pydantic_core.ValidationError: If type validation of arguments fails.
        """
        converters = self._replace_default_converters(query_case, body_case, cookie_case, header_case, response_case)

        decorator = self._get_decorator(
            path=path,
            method="PATCH",
            case_converters=converters,
            skip_finalizer=skip_finalizer
        )
        return decorator

    def put(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            response_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> _RouteDecorator:
        """
        Create a route using the HTTP PUT method.

        This decorator transforms a function into a routed function that sends a PUT request
        to the specified path, handling parameter case conversion and argument validation.

        Args:
            path (str):
                The relative path of the route.
            query_case (CaseConverter | None, optional):
                Case converter for query parameters. Defaults to using the router's default.
            body_case (CaseConverter | None, optional):
                Case converter for body parameters. Defaults to using the router's default.
            cookie_case (CaseConverter | None, optional):
                Case converter for cookie parameters. Defaults to using the router's default.
            header_case (CaseConverter | None, optional):
                Case converter for header parameters. Defaults to using the router's default.
            response_case:
                Case converter for JSON response. Defaults to using the router's default.
            skip_finalizer:
                Skip JSON finalizer, when finalizing response. Defaults to `False`.

        Returns:
            RoutedFunction: A routed function that sends a PUT request according to the specified path and validates
            its arguments based on type annotations.

        Raises:
            pydantic_core.ValidationError: If type validation of arguments fails.
        """
        converters = self._replace_default_converters(query_case, body_case, cookie_case, header_case, response_case)

        decorator = self._get_decorator(
            path=path,
            method="PUT",
            case_converters=converters,
            skip_finalizer=skip_finalizer
        )
        return decorator

    def delete(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            response_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> _RouteDecorator:
        """
        Create a route using the HTTP DELETE method.

        This decorator transforms a function into a routed function that sends a DELETE request
        to the specified path, handling parameter case conversion and argument validation.

        Args:
            path (str):
                The relative path of the route.
            query_case (CaseConverter | None, optional):
                Case converter for query parameters. Defaults to using the router's default.
            body_case (CaseConverter | None, optional):
                Case converter for body parameters. Defaults to using the router's default.
            cookie_case (CaseConverter | None, optional):
                Case converter for cookie parameters. Defaults to using the router's default.
            header_case (CaseConverter | None, optional):
                Case converter for header parameters. Defaults to using the router's default.
            response_case:
                Case converter for JSON response. Defaults to using the router's default.
            skip_finalizer:
                Skip JSON finalizer, when finalizing response. Defaults to `False`.

        Returns:
            RoutedFunction: A routed function that sends a DELETE request according to the specified path and validates
            its arguments based on type annotations.

        Raises:
            pydantic_core.ValidationError: If type validation of arguments fails.
        """
        converters = self._replace_default_converters(query_case, body_case, cookie_case, header_case, response_case)

        decorator = self._get_decorator(
            path=path,
            method="DELETE",
            case_converters=converters,
            skip_finalizer=skip_finalizer
        )
        return decorator



import asyncio
from abc import ABC, abstractmethod
import threading
from time import time, sleep
from sensei.types import IRateLimit


class RateLimit(IRateLimit):
    """
    A class that manages rate limiting by maintaining tokens and enforcing rate limits.

    Args:
        calls (int): The maximum number of requests allowed per period.
        period (int): The time period in seconds for the rate limit.
    """

    __slots__ = "_tokens", "_last_checked", "_async_lock", "_thread_lock"

    def __init__(self, calls: int, period: int) -> None:
        super().__init__(calls, period)
        self._tokens: int = calls  # Initial number of tokens equals rate_limit
        self._last_checked: float = time()  # Timestamp of the last check
        self._async_lock: asyncio.Lock = asyncio.Lock()  # Lock for synchronizing access
        self._thread_lock: threading.Lock = threading.Lock()

    def _acquire(self) -> bool:
        now: float = time()
        elapsed: float = now - self._last_checked

        # Replenish tokens based on elapsed time
        self._tokens += int(elapsed / self._period * self._calls)
        self._tokens = min(self._tokens, self._calls)  # Ensure tokens don't exceed rate_limit
        self._last_checked = now

        if self._tokens > 0:
            self._tokens -= 1  # Consume a token
            return True
        else:
            return False

    async def async_acquire(self) -> bool:
        """
        Asynchronously attempt to acquire a token.

        Returns:
            bool: True if a token was acquired, False otherwise.
        """
        async with self._async_lock:
            return self._acquire()

    async def async_wait_for_slot(self) -> None:
        """Asynchronously wait until a slot becomes available by periodically attempting to acquire a token."""
        while not await self.async_acquire():
            await asyncio.sleep(self._period / self._calls)

    def acquire(self) -> bool:
        """
        Synchronously attempt to acquire a token.

        Returns:
            bool: True if a token was acquired, False otherwise.
        """
        with self._thread_lock:
            return self._acquire()

    def wait_for_slot(self) -> None:
        """Synchronously wait until a slot becomes available by periodically attempting to acquire a token."""
        while not self.acquire():
            # Wait for the rate limit period before trying again
            sleep(self._period / self._calls)


class _BaseLimiter(ABC):
    def __init__(self, rate_limit: IRateLimit) -> None:
        self._rate_limit: IRateLimit = rate_limit

    @abstractmethod
    def acquire(self) -> bool:
        pass

    @abstractmethod
    def wait_for_slot(self) -> None:
        pass


class AsyncRateLimiter(_BaseLimiter):
    """
    Asynchronous rate limiter that manages request rate limiting using async methods.

    Args:
        rate_limit (IRateLimit): An instance of RateLimit to share between limiters.
    """

    def __init__(self, rate_limit: IRateLimit) -> None:
        super().__init__(rate_limit)

    async def acquire(self) -> bool:
        """
        Asynchronously attempt to acquire a token.

        Returns:
            bool: True if a token was acquired, False otherwise.
        """
        return await self._rate_limit.async_acquire()

    async def wait_for_slot(self) -> None:
        """Asynchronously wait until a slot becomes available by periodically acquiring a token."""
        await self._rate_limit.async_wait_for_slot()


class RateLimiter(_BaseLimiter):
    """
    Synchronous rate limiter that manages request rate limiting using synchronous methods.

    Args:
        rate_limit (IRateLimit): An instance of RateLimit to share between limiters.
    """

    def __init__(self, rate_limit: IRateLimit) -> None:
        super().__init__(rate_limit)

    def acquire(self) -> bool:
        """
        Synchronously attempt to acquire a token.

        Returns:
            bool: True if a token was acquired, False otherwise.
        """
        return self._rate_limit.acquire()

    def wait_for_slot(self) -> None:
        """Synchronously wait until a slot becomes available by periodically acquiring a token."""
        self._rate_limit.wait_for_slot()



import functools
from typing import Any
from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass
from ._internal import RoutedMethod, Hook, Args
from ._utils import bind_attributes
from ._utils import is_staticmethod, is_classmethod, is_selfmethod
from .types import Json


class _Namespace(dict):
    @staticmethod
    def _decorate_method(method: RoutedMethod) -> None:
        try:
            preparer = method.__func__.prepare
            finalizer = method.__func__.finalize

            bind_attributes(method, finalizer, preparer)  # type: ignore
        except AttributeError:
            pass

    def __setitem__(self, key: Any, value: Any):
        if is_staticmethod(value) or is_classmethod(value):
            self._decorate_method(value)
        elif key in Hook.values() and is_selfmethod(value):
            value = functools.partial(value, None)

        super().__setitem__(key, value)


class _ModelMeta(ModelMetaclass):
    @classmethod
    def __prepare__(metacls, name, bases):
        return _Namespace()

    def __new__(
            cls,
            cls_name: str,
            bases: tuple[type[Any], ...],
            namespace: dict[str, Any],
    ):
        obj = super().__new__(cls, cls_name, bases, namespace)
        obj.__router__ = None

        return obj


class APIModel(BaseModel, metaclass=_ModelMeta):
    """
    Base class for creating Sensei API models. Don`t confuse it with Pydantic BaseModel, it`s used simultaneously
    for validating output data and provide "routed" functions.

    But usage rools are the same with BaseModel
    Usage docs: https://docs.pydantic.dev/2.9/concepts/models/

    To make the proper class, decorate it with @router.model. Avoiding this requirements will lead you to the issues.

    Examples:
        >>> from typing import Annotated, Any, Self
        >>> from sensei import Router, Query, Path, APIModel
        ...
        >>> router = Router('https://reqres.in/api')
        ...
        >>> @router.model()
        ... class User(APIModel):
        ...     email: str
        ...     id: int
        ...     first_name: str
        ...     last_name: str
        ...     avatar: str
        ...
        ...     @classmethod
        ...     @router.get('/users/{id_}')
        ...     def get(cls, id_: Annotated[int, Path(alias='id')]) -> Self:
        ...         ...
    """

    @staticmethod
    def __finalize_json__(json: Json) -> Json:
        """
        Finalize the JSON response.

        This hook is used to finalize the JSON response. The final value must be JSON serializable.

        Args:
            json (Json): The original JSON response.

        Returns:
            dict[str, Any]: The finalized JSON response.
        """
        return json

    @staticmethod
    def __prepare_args__(args: Args) -> Args:
        """
        Prepare the arguments for the request.

        This hook is used to prepare the arguments for the request before it is sent.
        The final value must be an instance of `Args`.

        Args:
            args (Args): The original arguments.

        Returns:
            Args: The prepared arguments.
        """
        return args

    @staticmethod
    def __query_case__(s: str) -> str:
        """
        Convert the case of query parameters.

        This hook is used to convert the case of query parameters to the desired format.

        Args:
            s (str): The original query parameter string.

        Returns:
            str: The converted query parameter string.
        """
        return s

    @staticmethod
    def __body_case__(s: str) -> str:
        """
        Convert the case of body parameters.

        This hook is used to convert the case of body parameters to the desired format.

        Args:
            s (str): The original body parameter string.

        Returns:
            str: The converted body parameter string.
        """
        return s

    @staticmethod
    def __cookie_case__(s: str) -> str:
        """
        Convert the case of cookie parameters.

        This hook is used to convert the case of cookie parameters to the desired format.

        Args:
            s (str): The original cookie parameter string.

        Returns:
            str: The converted cookie parameter string.
        """
        return s

    @staticmethod
    def __header_case__(s: str) -> str:
        """
        Convert the case of header parameters.

        This hook is used to convert the case of header parameters to the desired format.

        Args:
            s (str): The original header parameter string.

        Returns:
            str: The converted header parameter string.
        """
        return s

    @staticmethod
    def __response_case__(s: str) -> str:
        """
        Convert the case of the JSON response keys.

        This hook is used to convert the case of the JSON response keys to the desired format.

        Args:
            s (str): The original JSON key string.

        Returns:
            str: The converted JSON key string.
        """
        return s

    def __str__(self):
        return f'{self.__class__.__name__}({super().__str__()})'



from typing import Any


class CollectionLimitError(ValueError):
    def __init__(self, collection: Any, limit: int, element_name: str = "element"):
        super().__init__(
            f'{collection} size limit exceeded. '
            f'It can contain only {limit} {element_name.lower()}{"s" if limit != 1  else ""}.'
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Protocol, TypeVar
from ._endpoint import CaseConverter
from ._requester import ResponseFinalizer, Preparer, JsonFinalizer
from ..tools import MethodType


class RoutedFunction(Protocol):
    def __call__(self, *args, **kwargs):
        ...

    def prepare(self, preparer: Preparer) -> Preparer:
        ...

    def finalize(self, finalizer: ResponseFinalizer) -> ResponseFinalizer:
        ...

    __method_type__: MethodType
    __name__: str
    __doc__: str


class RoutedMethod(Protocol):
    __func__: RoutedFunction
    __name__: str
    __doc__: str


class _RoutedModel(Protocol):
    __router__ = ...
    __finalize_json__: JsonFinalizer
    __prepare_args__: Preparer
    __query_case__: CaseConverter
    __body_case__: CaseConverter
    __cookie_case__: CaseConverter
    __header_case__: CaseConverter


RoutedModel = TypeVar("RoutedModel", bound=_RoutedModel)


class IRouter(ABC):
    __slots__ = ()

    @abstractmethod
    def model(self, model_obj: RoutedModel | None) -> RoutedModel:
        pass

    @abstractmethod
    def get(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> RoutedFunction:
        pass

    @abstractmethod
    def post(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> RoutedFunction:
        pass

    @abstractmethod
    def patch(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> RoutedFunction:
        pass

    @abstractmethod
    def put(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> RoutedFunction:
        pass

    @abstractmethod
    def delete(
            self,
            path: str,
            /, *,
            query_case: CaseConverter | None = None,
            body_case: CaseConverter | None = None,
            cookie_case: CaseConverter | None = None,
            header_case: CaseConverter | None = None,
            skip_finalizer: bool = False,
    ) -> RoutedFunction:
        pass



import inspect
import re
from typing import Any, TypeVar, Protocol

_T = TypeVar("_T")


class _NamedObj(Protocol):
    __name__ = ...


def is_classmethod(obj: Any) -> bool:
    return isinstance(obj, classmethod)


def is_staticmethod(obj: Any) -> bool:
    return isinstance(obj, staticmethod)


def is_instancemethod(obj: Any) -> bool:
    return inspect.isfunction(obj)


def is_selfmethod(obj: Any) -> bool:
    return is_classmethod(obj) or is_instancemethod(obj)


def is_method(obj: Any) -> bool:
    return is_selfmethod(obj) or is_staticmethod(obj)


def bind_attributes(obj: _T, *named_objects: tuple[_NamedObj]) -> _T:
    for named_obj in named_objects:
        setattr(obj, named_obj.__name__, named_obj)

    return obj


def placeholders(url: str) -> list[str]:
    """
    Extracts placeholder names from a string.

    This function searches the string for placeholders in the format `{param_name}`
    and returns a list of all parameter names found in the string.

    Args:
        url (str): The string containing placeholders in the format `{param_name}`.

    Returns:
        list[str]: A list of placeholder names found in the string. Each name is a string
                   corresponding to the `param_name` inside `{}` brackets.

    Example:
        >>> url = "https://example.com/users/{user_id}/posts/{post_id}"
        >>> result = placeholders(url)
        ["user_id", "post_id"]
    """
    pattern = r'\{(\w+)\}'

    parameters = re.findall(pattern, url)

    return parameters


def format_str(s: str, values: dict[str, Any], ignore_missed: bool = False) -> str:
    """
    Replaces placeholders in the string with corresponding values from the provided dictionary.

    This function searches the string for placeholders in the format `{param_name}` and replaces
    them with the value from the `values` dictionary where the key is `param_name`. If no corresponding
    value is found for a placeholder, it raises KeyError

    Args:
        s (str): String containing placeholders in the format `{param_name}`.
        values (dict[str, Any]): Dictionary where keys are parameter names and values are
                                 used to replace the placeholders in the URL.
        ignore_missed: Whether to ignore missed values

    Returns:
        str: String with placeholders replaced by corresponding values from the `values` dictionary.

    Raises
        KeyError: If no corresponding value is found for a placeholder

    Example:
        >>> url = "https://example.com/users/{user_id}/posts/{post_id}"
        >>> values = {"user_id": 42, "post_id": 1001}
        >>> format_str(url, values)
        https://example.com/users/42/posts/1001
    """
    try:
        return s.format(**values)
    except KeyError:
        if not ignore_missed:
            raise
        else:
            keys = placeholders(s)
            values = values | {k: f'{"{"+ k +"}"}' for k in keys}
            return s.format(**values)


def get_base_url(host: str, port: int) -> str:
    if 'port' in placeholders(host):
        api_url = format_str(host, {'port': port})
    elif port is not None:
        api_url = f'{host}:{port}'
    else:
        api_url = host

    return api_url



from collections.abc import Mapping
from typing import Iterator, TypeVar, Any, Hashable

_KT = TypeVar('_KT')
_VT = TypeVar('_VT')


class ChainedMap(Mapping[_KT, _VT]):
    def __init__(self, *dicts: dict[Any, Any]) -> None:
        self._dicts: tuple[dict[Any, Any], ...] = dicts

    def __getitem__(self, key: _KT) -> _VT:
        for d in self._dicts:
            if key in d:
                value = d[key]
                if any(value in next_d for next_d in self._dicts if isinstance(value, Hashable) and next_d is not d):
                    return self.__getitem__(value)
                return value
        raise KeyError(key)

    def __iter__(self) -> Iterator[Any]:
        for key in self._dicts[0]:
            yield from self.trace(key)

    def __len__(self) -> int:
        unique_keys = {key for d in self._dicts for key in d}
        return len(unique_keys)

    def trace(self, key: _KT) -> Iterator[Any]:
        """Yields the key and follows its chain through the dictionaries."""
        seen = []

        current_value = key
        while True:
            if current_value in seen:
                break
            seen.append(current_value)
            yield current_value
            for d in self._dicts:
                try:
                    if current_value in d:
                        current_value = d[current_value]
                        break
                except TypeError:
                    pass
            else:
                break



from __future__ import annotations

import inspect
from typing import Callable, TypeVar, Generic, Any, get_origin, get_args
from typing_extensions import Self
from sensei.client import Manager, AsyncClient, Client
from sensei._base_client import BaseClient
from ._endpoint import Endpoint, ResponseModel, RESPONSE_TYPES, CaseConverter, Args
from ._requester import Requester, ResponseFinalizer, Preparer, JsonFinalizer
from ..tools import HTTPMethod, args_to_kwargs, MethodType, identical
from sensei.types import IRateLimit, IResponse
from ..tools.utils import is_coroutine_function

_Client = TypeVar('_Client', bound=BaseClient)
_RequestArgs = tuple[tuple[Any, ...], dict[str, Any]]


class _CallableHandler(Generic[_Client]):
    __slots__ = (
        '_func',
        '_manager',
        '_method',
        '_path',
        '_rate_limit',
        '_host',
        '_port',
        '_request_args',
        '_method_type',
        '_temp_client',
        '_response_finalizer',
        '_preparer',
        '_converters',
        '_json_finalizer',
    )

    def __init__(
            self,
            *,
            path: str,
            method: HTTPMethod,
            func: Callable,
            host: str,
            port: int | None = None,
            rate_limit: IRateLimit | None = None,
            request_args: _RequestArgs,
            method_type: MethodType,
            manager: Manager[_Client] | None,
            case_converters: dict[str, CaseConverter],
            response_finalizer: ResponseFinalizer | None = None,
            json_finalizer: JsonFinalizer = identical,
            pre_preparer: Preparer = identical,
            post_preparer: Preparer = identical,
    ):
        self._func = func
        self._manager = manager
        self._method = method
        self._path = path
        self._host = host
        self._port = port
        self._rate_limit = rate_limit

        self._request_args = request_args
        self._method_type = method_type
        self._temp_client: _Client | None = None
        self._converters = case_converters

        if is_coroutine_function(post_preparer):
            async def preparer(value: Args) -> Args:
                return await post_preparer(pre_preparer(value))
        else:
            def preparer(value: Args) -> Args:
                return post_preparer(pre_preparer(value))

        if response_finalizer:
            if is_coroutine_function(response_finalizer):
                async def finalizer(value: IResponse) -> ResponseModel:
                    return await response_finalizer(value)
            else:
                def finalizer(value: IResponse) -> ResponseModel:
                    return response_finalizer(value)
        else:
            finalizer = response_finalizer

        self._preparer = preparer
        self._response_finalizer = finalizer

        self._json_finalizer = json_finalizer

    def __make_endpoint(self) -> Endpoint:
        params = {}
        func = self._func
        method_type = self._method_type
        sig = inspect.signature(func)

        args = sig.parameters.values()

        skipped = False

        for param in args:
            if MethodType.self_method(method_type) and not skipped:
                skipped = True
                continue

            if param.default and param.default is not inspect.Parameter.empty:
                params[param.name] = param.annotation, param.default
            else:
                params[param.name] = param.annotation

        return_type = sig.return_annotation if sig.return_annotation is not inspect.Signature.empty else None

        if not Endpoint.is_response_type(return_type):
            if return_type is Self:
                if MethodType.self_method(method_type):
                    return_type = func.__self__  # type: ignore
                else:
                    raise ValueError('Response "Self" is only for instance and class methods')
            elif get_origin(return_type) is list and get_args(return_type)[0] is Self:
                if method_type is MethodType.CLASS:
                    return_type = list[func.__self__]  # type: ignore
                else:
                    raise ValueError('Response "list[Self]" is only for class methods')
            elif self._response_finalizer is None:
                raise ValueError(f'Response finalizer must be set, if response is not from: {RESPONSE_TYPES}')
            else:
                return_type = dict

        converters = self._converters.copy()
        converters.pop('response_case')
        endpoint = Endpoint(self._path, self._method, params=params, response=return_type, **converters)
        return endpoint

    def _make_requester(self, client: BaseClient) -> Requester:
        case_converter = self._converters.get('response_case', identical)
        endpoint = self.__make_endpoint()
        requester = Requester(
            client,
            endpoint,
            response_finalizer=self._response_finalizer,
            json_finalizer=self._json_finalizer,
            preparer=self._preparer,
            response_case=case_converter,
        )
        return requester

    def _get_request_args(self, client: BaseClient) -> tuple[Requester, dict]:
        if client.host != self._host:
            raise ValueError('Client host must be equal to default host')

        if client.port != self._port:
            raise ValueError('Client port must be equal to default port')

        requester = self._make_requester(client)
        kwargs = args_to_kwargs(self._func, *self._request_args[0], **self._request_args[1])
        method_type = self._method_type

        if MethodType.self_method(method_type):
            kwargs.popitem(False)

        return requester, kwargs


class AsyncCallableHandler(_CallableHandler[AsyncClient], Generic[ResponseModel]):
    async def __aenter__(self) -> ResponseModel:
        manager = self._manager
        if manager is None or manager.empty():
            client = AsyncClient(host=self._host, port=self._port, rate_limit=self._rate_limit)
            await client.__aenter__()
            self._temp_client = client
        else:
            client = manager.get()
            if not isinstance(client, AsyncClient):
                raise TypeError(f'Manager`s client must be type of {AsyncClient}')

        requester, kwargs = self._get_request_args(client)

        return await requester.request(**kwargs)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if client := self._temp_client:
            await client.__aexit__(exc_type, exc_val, exc_tb)
            self._temp_client = None


class CallableHandler(_CallableHandler[Client], Generic[ResponseModel]):
    def __enter__(self) -> ResponseModel:
        manager = self._manager
        if manager is None or manager.empty():
            client = Client(host=self._host, port=self._port, rate_limit=self._rate_limit)
            client.__enter__()
            self._temp_client = client
        else:
            client = manager.get()
            if not isinstance(client, Client):
                raise TypeError(f'Manager`s client must be type of {Client}')

        requester, kwargs = self._get_request_args(client)

        return requester.request(**kwargs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if client := self._temp_client:
            client.__exit__(exc_type, exc_val, exc_tb)
            self._temp_client = None



from pydantic.version import VERSION as P_VERSION

PYDANTIC_VERSION = P_VERSION
PYDANTIC_V2 = PYDANTIC_VERSION.startswith("2.")

if PYDANTIC_V2:
    from pydantic_core import PydanticUndefined
    Undefined = PydanticUndefined
else:
    from pydantic.fields import Undefined



from typing import Any
from sensei.types import IRateLimit


class _Attr:
    def __set_name__(self, owner: type, name: str) -> None:
        self._name = name

    def __get__(self, obj: object, owner: type) -> Any:
        return obj.__dict__[self._name]


class RateLimitAttr(_Attr):
    def __get__(self, obj: object, owner: type) -> IRateLimit:
        return super().__get__(obj, owner)

    def __set__(self, obj: object, value: IRateLimit) -> None:
        if value is None or isinstance(value, IRateLimit):
            obj.__dict__[self._name] = value
        else:
            raise TypeError(f'Value must implement {IRateLimit} interface')


class PortAttr(_Attr):
    def __get__(self, obj: object, owner: type) -> int:
        return super().__get__(obj, owner)

    def __set__(self, obj: object, value: int) -> None:
        if value is None or isinstance(value, int) and 1 <= value <= 65535:
            obj.__dict__[self._name] = value
        else:
            raise ValueError('Port must be between 1 and 65535')



from __future__ import annotations

import typing
from httpx import Client as _Client, AsyncClient as _AsyncClient, USE_CLIENT_DEFAULT, Response, BaseTransport
from httpx._client import UseClientDefault, EventHook
from httpx._config import DEFAULT_MAX_REDIRECTS, Limits, DEFAULT_LIMITS, DEFAULT_TIMEOUT_CONFIG
from httpx._types import (URLTypes, RequestContent, RequestData, RequestFiles, QueryParamTypes, HeaderTypes,
                          CookieTypes, AuthTypes, TimeoutTypes, RequestExtensions, ProxiesTypes, ProxyTypes, CertTypes,
                          VerifyTypes)
from sensei._base_client import BaseClient
from .rate_limiter import IRateLimit, RateLimiter, AsyncRateLimiter


class Client(_Client, BaseClient):
    """An HTTP client, with connection pooling, HTTP/2, redirects, cookie persistence, etc."""

    def __init__(
            self,
            *,
            host: str,
            port: int | None = None,
            rate_limit: IRateLimit | None = None,
            auth: AuthTypes | None = None,
            params: QueryParamTypes | None = None,
            headers: HeaderTypes | None = None,
            cookies: CookieTypes | None = None,
            verify: VerifyTypes = True,
            cert: CertTypes | None = None,
            http1: bool = True,
            http2: bool = False,
            proxy: ProxyTypes | None = None,
            proxies: ProxiesTypes | None = None,
            mounts: typing.Mapping[str, BaseTransport | None] | None = None,
            timeout: TimeoutTypes = DEFAULT_TIMEOUT_CONFIG,
            follow_redirects: bool = False,
            limits: Limits = DEFAULT_LIMITS,
            max_redirects: int = DEFAULT_MAX_REDIRECTS,
            event_hooks: typing.Mapping[str, list[EventHook]] | None = None,
            transport: BaseTransport | None = None,
            app: typing.Callable[..., typing.Any] | None = None,
            trust_env: bool = True,
            default_encoding: str | typing.Callable[[bytes], str] = "utf-8",
    ) -> None:
        BaseClient.__init__(self, host=host, port=port, rate_limit=rate_limit)
        _Client.__init__(
            self,
            auth=auth,
            params=params,
            headers=headers,
            cookies=cookies,
            verify=verify,
            cert=cert,
            http1=http1,
            http2=http2,
            proxy=proxy,
            proxies=proxies,
            mounts=mounts,
            timeout=timeout,
            follow_redirects=follow_redirects,
            limits=limits,
            max_redirects=max_redirects,
            event_hooks=event_hooks,
            transport=transport,
            app=app,
            default_encoding=default_encoding,
            trust_env=trust_env,
            base_url=self._api_url
        )

    def request(
            self,
            method: str,
            url: URLTypes,
            *,
            content: RequestContent | None = None,
            data: RequestData | None = None,
            files: RequestFiles | None = None,
            json: typing.Any | None = None,
            params: QueryParamTypes | None = None,
            headers: HeaderTypes | None = None,
            cookies: CookieTypes | None = None,
            auth: AuthTypes | UseClientDefault | None = USE_CLIENT_DEFAULT,
            follow_redirects: bool | UseClientDefault = USE_CLIENT_DEFAULT,
            timeout: TimeoutTypes | UseClientDefault = USE_CLIENT_DEFAULT,
            extensions: RequestExtensions | None = None,
    ) -> Response:
        """Build and send an HTTP request."""
        if self.rate_limit:
            RateLimiter(self.rate_limit).wait_for_slot()

        return super().request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions
        )


class AsyncClient(_AsyncClient, BaseClient):
    """An HTTP client, with connection pooling, HTTP/2, redirects, cookie persistence, etc."""

    def __init__(
            self,
            *,
            host: str,
            port: int | None = None,
            rate_limit: IRateLimit | None = None,
            auth: AuthTypes | None = None,
            params: QueryParamTypes | None = None,
            headers: HeaderTypes | None = None,
            cookies: CookieTypes | None = None,
            verify: VerifyTypes = True,
            cert: CertTypes | None = None,
            http1: bool = True,
            http2: bool = False,
            proxy: ProxyTypes | None = None,
            proxies: ProxiesTypes | None = None,
            mounts: typing.Mapping[str, BaseTransport | None] | None = None,
            timeout: TimeoutTypes = DEFAULT_TIMEOUT_CONFIG,
            follow_redirects: bool = False,
            limits: Limits = DEFAULT_LIMITS,
            max_redirects: int = DEFAULT_MAX_REDIRECTS,
            event_hooks: typing.Mapping[str, list[EventHook]] | None = None,
            transport: BaseTransport | None = None,
            app: typing.Callable[..., typing.Any] | None = None,
            trust_env: bool = True,
            default_encoding: str | typing.Callable[[bytes], str] = "utf-8",
    ) -> None:
        BaseClient.__init__(self, host=host, port=port, rate_limit=rate_limit)
        _AsyncClient.__init__(
            self,
            auth=auth,
            params=params,
            headers=headers,
            cookies=cookies,
            verify=verify,
            cert=cert,
            http1=http1,
            http2=http2,
            proxy=proxy,
            proxies=proxies,
            mounts=mounts,
            timeout=timeout,
            follow_redirects=follow_redirects,
            limits=limits,
            max_redirects=max_redirects,
            event_hooks=event_hooks,
            transport=transport,
            app=app,
            default_encoding=default_encoding,
            trust_env=trust_env,
            base_url=self._api_url
        )

    async def request(
            self,
            method: str,
            url: URLTypes,
            *,
            content: RequestContent | None = None,
            data: RequestData | None = None,
            files: RequestFiles | None = None,
            json: typing.Any | None = None,
            params: QueryParamTypes | None = None,
            headers: HeaderTypes | None = None,
            cookies: CookieTypes | None = None,
            auth: AuthTypes | UseClientDefault | None = USE_CLIENT_DEFAULT,
            follow_redirects: bool | UseClientDefault = USE_CLIENT_DEFAULT,
            timeout: TimeoutTypes | UseClientDefault = USE_CLIENT_DEFAULT,
            extensions: RequestExtensions | None = None,
    ) -> Response:
        """Build and send an HTTP request."""
        if self.rate_limit:
            await AsyncRateLimiter(self.rate_limit).wait_for_slot()

        return await super().request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions
        )



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Protocol, Mapping, Any, Union
from typing_extensions import Self

Json = Union[dict, list[dict]]


class IRateLimit(ABC):
    __slots__ = "_calls", "_period"

    def __init__(self, calls: int, period: int) -> None:
        """
        Initialize the shared state for rate limiting.

        :param calls: Maximum number of requests allowed per period.
        :param period: Time period (in seconds) for the rate limit.
        """
        self._calls: int = calls
        self._period: int = period

    @property
    def period(self) -> int:
        return self._period

    @period.setter
    def period(self, period: int) -> None:
        self._period = period

    @property
    def calls(self):
        return self._calls

    @calls.setter
    def calls(self, rate_limit: int) -> None:
        self._calls = rate_limit

    @abstractmethod
    async def async_acquire(self) -> bool:
        """
        Attempt to acquire a token.

        :return: True if a token was acquired, False otherwise.
        """
        pass

    @abstractmethod
    async def async_wait_for_slot(self) -> None:
        """
        Wait until a slot becomes available by periodically acquiring a token.
        """
        pass

    @abstractmethod
    def acquire(self) -> bool:
        """
        Synchronously attempt to acquire a token.

        :return: True if a token was acquired, False otherwise.
        """
        pass

    @abstractmethod
    def wait_for_slot(self) -> None:
        """
        Wait until a slot becomes available by periodically acquiring a token.
        """
        pass


class IRequest(Protocol):
    @property
    def headers(self) -> Mapping[str, Any]:
        pass

    @property
    def method(self) -> str:
        pass

    @property
    def url(self) -> Any:
        pass


class IResponse(Protocol):
    __slots__ = ()

    def __await__(self):
        pass

    @abstractmethod
    def json(self) -> Json:
        pass

    @abstractmethod
    def raise_for_status(self) -> Self:
        pass

    @property
    @abstractmethod
    def request(self) -> IRequest:
        pass

    @property
    @abstractmethod
    def text(self) -> str:
        pass

    @property
    @abstractmethod
    def status_code(self) -> int:
        pass

    @property
    @abstractmethod
    def content(self) -> bytes:
        pass



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from typing import Callable, Generic, Any, Awaitable, Union
from ._endpoint import Endpoint, Args, ResponseModel, CaseConverter
from sensei._base_client import BaseClient
from sensei.client import Client, AsyncClient
from sensei.types import IResponse, IRequest, Json
from ..tools import identical
from sensei._utils import placeholders

Preparer = Callable[[Args], Union[Args, Awaitable[Args]]]
ResponseFinalizer = Callable[[IResponse], Union[ResponseModel, Awaitable[ResponseModel]]]
JsonFinalizer = Callable[[Json], Json]


class _DecoratedResponse(IResponse):
    __slots__ = (
        "_response",
        "_json_finalizer",
        "_response_case"
    )

    def __init__(
            self,
            response: IResponse,
            json_finalizer: JsonFinalizer = identical,
            response_case: CaseConverter = identical,
    ):
        self._response = response
        self._json_finalizer = json_finalizer
        self._response_case = response_case

    def json(self) -> Json:
        case = self._response_case
        json = self._response.json()
        json = {case(k): v for k, v in json.items()}
        return self._json_finalizer(json)

    def raise_for_status(self) -> IResponse:
        return self._response.raise_for_status()

    def request(self) -> IRequest:
        return self._response.request

    @property
    def text(self) -> str:
        return self._response.text

    @property
    def status_code(self) -> int:
        return self._response.status_code

    @property
    def content(self) -> bytes:
        return self._response.content


class Requester(ABC, Generic[ResponseModel]):
    __slots__ = (
        "_client",
        "_post_preparer",
        "_response_finalizer",
        "_endpoint",
        "_json_finalizer",
        "_preparer",
        "_is_async_preparer",
        "_is_async_response_finalizer",
        "_response_case"
    )

    def __new__(
            cls,
            client: BaseClient,
            endpoint: Endpoint,
            *,
            response_finalizer: ResponseFinalizer | None = None,
            json_finalizer: JsonFinalizer = identical,
            preparer: Preparer = identical,
            response_case: CaseConverter = identical,
    ):
        if isinstance(client, AsyncClient):
            return super().__new__(_AsyncRequester)
        elif isinstance(client, Client):
            return super().__new__(_Requester)
        else:
            raise ValueError("Client must be an instance of AsyncClient or Client")

    def __init__(
            self,
            client: BaseClient,
            endpoint: Endpoint,
            *,
            response_finalizer: ResponseFinalizer | None = None,
            json_finalizer: JsonFinalizer = identical,
            preparer: Preparer = identical,
            response_case: CaseConverter = identical,
    ):
        self._client = client
        self._response_finalizer = response_finalizer or self._finalize
        self._endpoint = endpoint
        self._json_finalizer = json_finalizer
        self._preparer = preparer
        self._is_async_preparer = inspect.iscoroutinefunction(self._preparer)
        self._is_async_response_finalizer = inspect.iscoroutinefunction(self._response_finalizer)
        self._response_case = response_case

    def _finalize(self, response: IResponse) -> ResponseModel:
        endpoint = self._endpoint
        return endpoint.get_response(response_obj=response)

    @abstractmethod
    def request(self, **kwargs) -> ResponseModel:
        pass

    def _dump_args(self, args: Args) -> dict[str, Any]:
        endpoint = self._endpoint
        args = args.model_dump(mode="json", exclude_none=True, by_alias=True)
        if placeholders(url := args['url']):
            raise ValueError(f'Path params of {url} params must be passed')
        return {'method': endpoint.method, **args}


class _AsyncRequester(Requester):
    async def _call_preparer(self, args: Args) -> Args:
        result = self._preparer(args)
        if self._is_async_preparer:
            result = await result
        return result

    async def _call_response_finalizer(self, response: IResponse) -> ResponseModel:
        result = self._response_finalizer(response)
        if self._is_async_response_finalizer:
            result = await result
        return result

    async def _get_args(self, **kwargs) -> dict[str, Any]:
        endpoint = self._endpoint
        args = endpoint.get_args(**kwargs)
        args = await self._call_preparer(args)
        return self._dump_args(args)

    async def request(self, **kwargs) -> ResponseModel:
        client = self._client
        args = await self._get_args(**kwargs)

        response = await client.request(**args)
        response.raise_for_status()
        response = _DecoratedResponse(response, json_finalizer=self._json_finalizer, response_case=self._response_case)
        return await self._call_response_finalizer(response)


class _Requester(Requester):
    def _call_preparer(self, args: Args) -> Args:
        result = self._preparer(args)
        if self._is_async_preparer:
            raise ValueError("If preparer is async, the route function must match it")
        return result

    def _call_response_finalizer(self, response: IResponse) -> ResponseModel:
        result = self._response_finalizer(response)
        if self._is_async_response_finalizer:
            raise ValueError("If response finalizer is async, the route function must match it")
        return result

    def _get_args(self, **kwargs) -> dict[str, Any]:
        endpoint = self._endpoint
        args = endpoint.get_args(**kwargs)
        args = self._call_preparer(args)
        return self._dump_args(args)

    def request(self, **kwargs) -> ResponseModel:
        client = self._client
        args = self._get_args(**kwargs)

        response = client.request(**args)
        response.raise_for_status()
        response = _DecoratedResponse(response, json_finalizer=self._json_finalizer, response_case=self._response_case)
        return self._call_response_finalizer(response)



from __future__ import annotations

from abc import abstractmethod, ABC
from ._utils import get_base_url
from .types import IRateLimit, IResponse
from sensei._descriptors import RateLimitAttr, PortAttr


class _PortAttr(PortAttr):
    def __set__(self, obj: object, value: int) -> None:
        if not obj.__dict__.get('_port_set'):
            super().__set__(obj, value)
            obj.__dict__['_port_set'] = True
        else:
            raise AttributeError('Port can only be set at creation')


class BaseClient(ABC):
    rate_limit = RateLimitAttr()
    port = _PortAttr()

    def __init__(
            self,
            host: str,
            port: int | None = None,
            rate_limit: IRateLimit | None = None,
    ):
        self.rate_limit = rate_limit
        self.port = port

        if host.endswith('/'):
            host = host[:-1]

        self._host = host

        base_url = get_base_url(host, port)
        self._api_url = base_url

    @property
    def host(self) -> str:
        return self._host

    @abstractmethod
    def request(self, method: str, url: str, *args, **kwargs) -> IResponse:
        pass

    @property
    @abstractmethod
    def base_url(self) -> str:
        pass



from ._internal import Router, Args
from .params import Path, Query, Cookie, Header, Body
from .client import RateLimit, Manager, Client, AsyncClient
from .api_model import APIModel
from .cases import *
from ._utils import format_str, placeholders
from .types import Json



import sys
import inspect
from functools import wraps
from pydantic import BaseModel
from collections import OrderedDict
from sensei._utils import placeholders
from typing import Any, get_args, Callable, TypeVar
from .types import HTTPMethod, MethodType
from pydantic._internal._model_construction import ModelMetaclass

_T = TypeVar("_T")


def make_model(model_name: str, model_args: dict[str, Any]) -> type[BaseModel]:
    annotations = {}
    defaults = {}
    for key, arg in model_args.items():
        if isinstance(arg, (tuple, list)):
            annotations[key] = arg[0]
            if len(arg) == 2:
                defaults[key] = arg[1]
        else:
            annotations[key] = arg

    model: type[BaseModel] = ModelMetaclass(  # type: ignore
        model_name,
        (BaseModel,),
        {
            '__module__': sys.modules[__name__],
            '__qualname__': model_name,
            '__annotations__': annotations,
            **defaults
        }
    )
    return model


def split_params(url: str, params: dict[str, Any]) -> tuple[dict[str, Any], dict[str, Any]]:
    path_params_names = placeholders(url)

    path_params = {}
    for path_param_name in path_params_names:
        if (value := params.get(path_param_name)) is not None:
            path_params[path_param_name] = value
            del params[path_param_name]

    return params, path_params


def is_safe_method(method: HTTPMethod) -> bool:
    if method in {'GET', 'TRACE', 'OPTIONS', 'HEAD'}:
        return True
    else:
        return False


def validate_method(method: HTTPMethod) -> bool:
    methods = get_args(HTTPMethod)
    if method not in methods:
        raise ValueError(f'Invalid HTTP method "{method}". '
                         f'Standard HTTP methods defined by the HTTP/1.1 protocol: {methods}')
    else:
        return True


def args_to_kwargs(func: Callable, *args, **kwargs) -> OrderedDict[str, Any]:
    sig = inspect.signature(func)

    bound_args = sig.bind_partial(*args, **kwargs)
    bound_args.apply_defaults()

    return OrderedDict(bound_args.arguments)


def set_method_type(func: Callable):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not args:
            method_type = MethodType.STATIC
        else:
            first_arg = args[0]

            func_name = getattr(first_arg, func.__name__, None)
            is_first_self = hasattr(first_arg, func.__name__) and getattr(func_name, '__self__', None) is first_arg

            if is_first_self:
                if inspect.isclass(first_arg):
                    method_type = MethodType.CLASS
                else:
                    method_type = MethodType.INSTANCE
            else:
                method_type = MethodType.STATIC

        setattr(wrapper, '__method_type__', method_type)

        return func(*args, **kwargs)

    return wrapper


def identical(value: _T) -> _T:
    return value


def is_coroutine_function(func: Callable) -> bool:
    return (inspect.iscoroutinefunction(func) or
            (hasattr(func, '__wrapped__') and inspect.iscoroutinefunction(func.__wrapped__)))



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from functools import wraps, partial
from typing import Callable, TypeVar
from sensei.client import Manager
from ._endpoint import CaseConverter
from ..tools import HTTPMethod, MethodType, identical
from sensei._base_client import BaseClient
from ._callable_handler import CallableHandler, AsyncCallableHandler
from ._requester import ResponseFinalizer, Preparer, JsonFinalizer
from sensei.types import IRateLimit

_Client = TypeVar('_Client', bound=BaseClient)


class Route(ABC):
    __slots__ = (
        '_path',
        '_method',
        '_func',
        '_manager',
        '_host',
        '_port',
        '_rate_limit',
        '_response_finalizer',
        '_method_type',
        '_is_async',
        '_case_converters',
        '_preparer',
        '_json_finalizer',
        '_pre_preparer',
        '__self__'
    )

    def __new__(
            cls,
            path: str,
            method: HTTPMethod,
            *,
            func: Callable,
            manager: Manager[_Client],
            host: str,
            port: int | None = None,
            rate_limit: IRateLimit | None = None,
            case_converters: dict[str, CaseConverter],
            json_finalizer: JsonFinalizer = identical,
            pre_preparer: Preparer = identical,
            response_case: CaseConverter = identical,
    ):
        if inspect.iscoroutinefunction(func):
            instance = super().__new__(_AsyncRoute)
            is_async = True
        else:
            instance = super().__new__(_SyncRoute)
            is_async = False

        instance._is_async = is_async

        return instance

    def __init__(
            self,
            path: str,
            method: HTTPMethod,
            *,
            func: Callable,
            manager: Manager[_Client],
            host: str,
            port: int | None = None,
            rate_limit: IRateLimit | None = None,
            case_converters: dict[str, CaseConverter],
            json_finalizer: JsonFinalizer | None = None,
            pre_preparer: Preparer = identical,
    ):
        self._path = path
        self._method = method
        self._func = func
        self._manager = manager
        self._rate_limit = rate_limit

        self._host = host
        self._port = port

        self._response_finalizer: ResponseFinalizer | None = None
        self._preparer: Preparer = identical
        self._method_type: MethodType = MethodType.STATIC

        self._case_converters = case_converters
        self._json_finalizer = json_finalizer
        self._pre_preparer = pre_preparer
        self.__self__: object | None = None

    @property
    def path(self) -> str:
        return self._path

    @property
    def method(self) -> HTTPMethod:
        return self._method

    @property
    def is_async(self) -> bool:
        return self._is_async

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass

    @property
    def method_type(self) -> MethodType:
        return self._method_type

    @method_type.setter
    def method_type(self, value: MethodType):
        if isinstance(value, MethodType):
            self._method_type = value
        else:
            raise TypeError(f'Method type must be an instance of {MethodType}')

    def _get_wrapper(self, func: Callable[..., ...]):
        @wraps(func)
        def wrapper(*args, **kwargs):
            new_func = func

            if self.__self__ is not None:
                new_func = partial(func, self.__self__)

            return new_func(*args, **kwargs)

        return wrapper

    def finalize(self, func: ResponseFinalizer | None = None) -> Callable:
        """
        Args:
            func (ResponseFinalizer | None):
                Response finalizer, used to modify final response, primarily when response type of "routed"
                function is not from declared. Executed after router`s __finalize_json__

        Returns:
            ResponseFinalizer: Wrapped function, used to finalize response
        """
        def decorator(func: ResponseFinalizer) -> ResponseFinalizer:
            self._response_finalizer = self._get_wrapper(func)
            return func

        if func is None:
            return decorator
        else:
            return decorator(func)

    def prepare(self, func: Preparer | None = None) -> Callable:
        """
        Args:
            func (Preparer | None):
                Args preparer, used to prepare the args for request before it. Final value also must be `Args` instance.
                Executed after router`s __prepare_args__

        Returns:
            Preparer: Wrapped function, used to prepare the args for request before it
        """
        def decorator(func: Preparer) -> Preparer:
            self._preparer = self._get_wrapper(func)
            return func

        if func is None:
            return decorator
        else:
            return decorator(func)


class _SyncRoute(Route):
    def __call__(self, *args, **kwargs):
        with CallableHandler(
            func=self._func,
            host=self._host,
            port=self._port,
            rate_limit=self._rate_limit,
            request_args=(args, kwargs),
            manager=self._manager,
            method_type=self._method_type,
            path=self.path,
            method=self._method,
            post_preparer=self._preparer,
            response_finalizer=self._response_finalizer,
            case_converters=self._case_converters,
            json_finalizer=self._json_finalizer,
            pre_preparer=self._pre_preparer
        ) as response:
            return response


class _AsyncRoute(Route):
    async def __call__(self, *args, **kwargs):
        async with AsyncCallableHandler(
            func=self._func,
            host=self._host,
            port=self._port,
            rate_limit=self._rate_limit,
            request_args=(args, kwargs),
            manager=self._manager,
            method_type=self._method_type,
            path=self.path,
            method=self._method,
            post_preparer=self._preparer,
            response_finalizer=self._response_finalizer,
            case_converters=self._case_converters,
            json_finalizer=self._json_finalizer,
            pre_preparer=self._pre_preparer
        ) as response:
            return response



from typing import TypeVar, Generic
from sensei._base_client import BaseClient
from .exceptions import CollectionLimitError

_ClientType = TypeVar('_ClientType', bound=BaseClient)


class Manager(Generic[_ClientType]):
    """
    A generic manager class for handling a single client instance with type safety. The client type defined at
    creation and it`s not reset.

    This class manages an instance of a client (which must be a subclass of `BaseClient`),
    ensuring that only one client is set at a time. The class provides methods to set,
    pop, and check the presence of the client.

    Args:
        client (_Client): An instance of a class that extends `BaseClient`.

    Raises:
        ValueError: If the provided client is not an instance of `BaseClient`.
    """

    __slots__ = ('_client', '_client_type')

    def __init__(self, client: _ClientType):
        if isinstance(client, BaseClient):
            self._client = client
            self._client_type = type(client)
        else:
            raise TypeError("Client must be an instance of AsyncClient or Client.")

    def set(self, client: _ClientType) -> None:
        """
        Set a client instance in the manager.

        This method sets the client if no client is currently set. If a client is
        already set, it raises a `CollectionLimitError`.

        Args:
            client (_Client): The client instance to set.

        Raises:
            CollectionLimitError: If a client is already set.
            TypeError: If the provided client is not an instance of the same type as the current client.
        """
        if isinstance(client, self._client_type):
            if self._client is None:
                self._client = client
            else:
                raise CollectionLimitError(self.__class__, 1, "Client")
        else:
            raise TypeError(f"Client must be an instance of {self._client_type.__name__}")

    def pop(self) -> _ClientType:
        """
        Remove and return the currently set client.

        This method removes the managed client and returns it. After calling this method,
        the manager will no longer have a client set. But the client type defined at creation is not reset.

        Returns:
            _Client: The client instance that was managed.

        Raises:
            AttributeError: If no client is set.
        """
        if self._client is None:
            raise AttributeError("Client is not set")
        else:
            client = self._client
            self._client = None
            return client

    def empty(self) -> bool:
        """
        Check if the manager has a client set.

        Returns:
            bool: True if no client is set, False otherwise.
        """
        return self._client is None

    def get(self) -> _ClientType:
        """
        Retrieve the currently set client.

        This method returns the managed client without removing it.

        Returns:
            _Client: The client instance that is being managed.

        Raises:
            AttributeError: If no client is set.
        """
        if self._client is None:
            raise AttributeError("Client is not set")
        else:
            return self._client



from typing import Literal
from enum import Enum
from typing_extensions import Self

HTTPMethod = Literal[
    "GET",
    "POST",
    "PUT",
    "DELETE",
    "PATCH",
    "HEAD",
    "OPTIONS",
    "CONNECT",
    "TRACE"
]


class MethodType(Enum):
    INSTANCE = "instance"
    CLASS = "class"
    STATIC = "static"

    @classmethod
    def self_method(cls, method: Self) -> bool:
        return method in (cls.CLASS, cls.INSTANCE)
